;****************************************************************************
;* A20 Line Control Module                                                  *
;* Programmed by ALI                                                        *
;* Version 1.0.2                                   Last revision: 15-Dec-93 *
;****************************************************************************

;--------------------------- History ---------------------------
; Ver    Date       Who     Description
; 1.0.1  20-Feb-93  ALI     Written
; 1.0.2  15-Dec-93  ALI     Fixed 'Can't control A20' bug in the
;                           function InstallA20. The function
;                           InstallA20 now doesn't destroy the
;                           ES segment register.

Ideal                                   ;
NOSMART                                 ;
P386                                    ;

include "a20.inc"                       ;

        segment code word public use16 'CODE'
        assume  cs:code,ds:code

A20Handler      dw      0               ; Offset of the A20 Handler

;---------------------------------------------------------------------------;
;                                                                           ;
;  EnableA20 -  Enable the A20 line                                         ;
;                                                                           ;
;  Expects: None                                                            ;
;  Returns: NC if the A20 line is enabled, CY otherwise.                    ;
;  Regs:    AX, BX and Flags clobbered                                      ;
;                                                                           ;
;---------------------------------------------------------------------------;

        proc    EnableA20

        cmp     [cs:A20Handler],0       ; Does A20 handler installed?
        jnz     @@Call                  ;
        call    InstallA20              ;
        jc      @@Ret                   ;
@@Call: mov     ax,1                    ;
        call    [cs:A20Handler]         ; Call machine-specific A20 handler
@@Ret:  ret                             ;
        endp                            ;

;---------------------------------------------------------------------------;
;                                                                           ;
;  DisableA20 - Disable the A20 line                                        ;
;                                                                           ;
;  Expects: None                                                            ;
;  Returns: NC if the A20 line is disabled, CY otherwise.                   ;
;  Regs:    AX, BX and Flags are clobbered                                  ;
;                                                                           ;
;---------------------------------------------------------------------------;

        proc    DisableA20              ;

        cmp     [cs:A20Handler],0       ; Does A20 handler installed?
        jnz     @@Call                  ;
        call    InstallA20              ;
        jc      @@Ret                   ;
@@Call: xor     ax,ax                   ;
        call    [cs:A20Handler]         ; Call machine-specific A20 handler
@@Ret:  ret                             ;
        endp                            ;


;---------------------------------------------------------------------------;
;                                                                           ;
;  IsA20On - Returns the state of the A20 line                              ;
;                                                                           ;
;  Expects: None                                                            ;
;  Returns: NZ if the A20 line is enabled, ZR otherwise                     ;
;  Regs:    AX, CX, DI, SI and Flags clobbered                              ;
;                                                                           ;
;---------------------------------------------------------------------------;

LowMem  dd      000000080h              ; Set equal to 0000:0080
HighMem dd      0FFFF0090h              ; Set equal to FFFF:0090
;---------------------------------------;
        proc    IsA20On                 ;

        push    ds                      ;
        push    es                      ;

        lds     si,[cs:LowMem]          ; Compare the four words at 0000:0080
        les     di,[cs:HighMem]         ;   with the four at FFFF:0090
        mov     cx,4                    ;
        cld                             ;
        repe    cmpsw                   ; Are the two areas the same?
        or      cx,cx                   ; Set ZF if yes

        pop     es                      ;
        pop     ds                      ;
        ret                             ;
        endp                            ;

;****************************************************************************
;*                                                                          *
;* A20 Handler Section                                                      *
;*                                                                          *
;****************************************************************************

;---------------------------------------------------------------------------;
;                                                                           ;
;  AT_A20 - Enable/Disable the A20 line on non-PS/2 machines                ;
;                                                                           ;
;  Expects: AX = 0 for Disable, 1 for Enable                                ;
;  Returns: NC for success, CY otherwise                                    ;
;  Regs:    AX, CX and Flags clobbered                                      ;
;                                                                           ;
;---------------------------------------------------------------------------;

        proc    AT_A20                  ;

        or      ax,ax                   ;
        jz      @@Disable               ;

        call    Sync8042                ; Make sure the Keyboard Controller
        jnz     @@Err                   ; is Ready

        mov     al,0D1h                 ; Send D1h
        out     64h,al                  ;
        call    Sync8042                ;
        jnz     @@Err                   ;

        mov     al,0DFh                 ; Send DFh
        out     60h,al                  ;
        call    Sync8042                ;
        jnz     @@Err                   ;

        ; Wait for the A20 line to settle down (up to 20usecs)
        mov     al,0FFh                 ; Send FFh (Pulse Output Port NULL)
        out     64h,al                  ;
        call    Sync8042                ;
        jnz     @@Err                   ;
        jmp     @@Exit                  ;

@@Disable:
        call    Sync8042                ; Make sure the Keyboard Controller
        jnz     @@Err                   ; is Ready

        mov     al,0D1h                 ; Send D1h
        out     64h,al                  ;
        call    Sync8042                ;
        jnz     @@Err                   ;

        mov     al,0DDh                 ; Send DDh
        out     60h,al                  ;
        call    Sync8042                ;
        jnz     @@Err                   ;

        ; Wait for the A20 line to settle down (up to 20usecs)
        mov     al,0FFh                 ; Send FFh (Pulse Output Port NULL)
        out     64h,al                  ;
        call    Sync8042                ;

@@Exit: clc                             ;
        ret                             ;
@@Err:  stc                             ;
        ret                             ;
        endp                            ;
;---------------------------------------;
        proc    Sync8042                ;
        xor     cx,cx                   ;
@@InSync:                               ;
        in      al,64h                  ;
        and     al,2                    ;
        loopnz  @@InSync                ;
        ret                             ;
        endp                            ;

;--------------------------------------------------------------------------;
;                                                                          ;
;  PS2_A20 - Enable/Disable the A20 line on PS/2 machines                  ;
;                                                                          ;
;  Expects: AX = 0 for Disable, 1 for Enable                               ;
;  Returns: NC for success, CY otherwise                                   ;
;  Regs:    AX, CX and Flags clobbered                                     ;
;                                                                          ;
;--------------------------------------------------------------------------;

PS2_PORTA       equ     0092h           ;
PS2_A20BIT      equ     00000010b       ;
;---------------------------------------;
        proc    PS2_A20                 ;

        or      ax,ax                   ;
        jz      @@Disable               ;

        in      al,PS2_PORTA            ; Get the current A20 state
        test    al,PS2_A20BIT           ; Is A20 already on?
        jnz     @@Err                   ;

        or      al,PS2_A20BIT           ; Turn on the A20 line
        out     PS2_PORTA,al            ;

        xor     cx,cx                   ; Make sure we loop for awhile
@@IsItOn:
        in      al,PS2_PORTA            ; Loop until the A20 line comes on
        test    al,PS2_A20BIT           ;
        loopz   @@IsItOn                ;
        jz      @@Err                   ; Unable to turn on the A20 line
        jmp     @@Exit                  ;

@@Disable:                              ;
        in      al,PS2_PORTA            ; Get the current A20 state
        and     al,NOT PS2_A20BIT       ; Turn off the A20 line

        out     PS2_PORTA,al            ;

        xor     cx,cx                   ; Make sure we loop for awhile
@@IsItOff:                              ;
        in      al,PS2_PORTA            ; Loop until the A20 line goes off
        test    al,PS2_A20BIT           ;
        loopnz  @@IsItOff               ;
        jnz     @@Err                   ; Unable to turn off the A20 line

@@Exit: clc                             ;
        ret                             ;
@@Err:  stc                             ;
        ret                             ;
        endp                            ;

;---------------------------------------------------------------------------;
;                                                                           ;
;  HP_A20Handler - Enable/Disable the A20 line on HP Vectra machines        ;
;                                                                           ;
;  Expexts: AX = 0 for Disable, 1 for Enable                                ;
;  Returns: NC for success, CY otherwise                                    ;
;  Regs:    AX, CX and Flags clobbered                                      ;
;                                                                           ;
;---------------------------------------------------------------------------;

        proc    HP_A20                  ;

        or      ax,ax                   ;
        jz      @@Disable               ;

        call    HPSync8042              ; Make sure the Keyboard Controller
        jnz     @@Err                   ; is ready

        mov     al,0DFh                 ; Send DFh
        out     64h,al                  ;
        call    HPSync8042              ;
        jnz     @@Err                   ;

        mov     al,0DFh                 ; Send second DFh
        out     64h,al                  ;
        call    HPSync8042              ;
        jnz     @@Err                   ;
        jmp     @@Exit                  ;

@@Disable:
        call    HPSync8042              ; Make sure the Keyboard Controller
        jnz     @@Err                   ; is Ready

        mov     al,0DDh                 ; Send DDh
        out     64h,al                  ;
        call    HPSync8042              ;
        jnz     @@Err                   ;

        mov     al,0DDh                 ; Send second DDh
        out     64h,al                  ;
        call    HPSync8042              ;
        jnz     @@Err                   ;

@@Exit: clc                             ;
        ret                             ;
@@Err:  stc                             ;
        ret                             ;
        endp                            ;
;---------------------------------------;
        proc    HPSync8042              ;
        xor     cx,cx                   ;
@@InSync:                               ;
        in      al,64h                  ;
        and     al,2                    ;
        loopnz  @@InSync                ;
        ret                             ;
        endp                            ;

;--------------------------------------------------------------------------*
;                                                                           *
;  InstallA20 - Attempt to install the proper A20 Handler                           *
;                                                                           *
;  Expects: None                                                            *
;  Returns: None                                                            ;
;  Regs:    AX, CX, DI, SI, ES and Flags are clobbered                      *
;                                                                           *
;--------------------------------------------------------------------------*

        proc    InstallA20              ;

        push    es                      ;
        call    IsPS2Machine            ; Are we on a PS/2?
        jc      @@CheckHP               ;

        ; Yes, install the PS/2 A20 handler.
        mov     [cs:A20Handler],offset PS2_A20
        jmp     @@ret                   ;


@@CheckHP:
        call    IsHPMachine             ; Are we on a HP Vectra?
        jc      @@OnAT                  ;

        ; Yes, install the HP Vectra A20 handler.
        mov     [cs:A20Handler],offset HP_A20
        jmp     @@ret                   ;

@@OnAT: mov     [cs:A20Handler],offset AT_A20
        clc                             ; We suppose that is PC/AT
@@ret:  pop     es                      ;
        ret                             ;
        endp                            ;

;---------------------------------------------------------------------------;
;                                                                           ;
;  IsPS2Machine - Check for PS/2 machine                                    ;
;                                                                           ;
;  Expects: None                                                            ;
;  Returns: NC if we're on a valid PS/2 machine, CY otherwise               ;
;  Regs:    ES, AX, BX and Flags clobbered                                  ;
;                                                                           ;
;---------------------------------------------------------------------------;

        proc    IsPS2Machine

        mov     ah,0C0h                 ; Get System Description Vector
        stc                             ;
        int     15h                     ;
        jc      @@NoPS2                 ; Error?  Not a PS/2.

        ; Do we have a "Micro Channel" computer?
        mov     al,[byte es:bx+5]       ; Get "Feature Information Byte 1"
        shr     al,2                    ; Test the "Micro Channel
        cmc                             ;
;       jc      @@NoPS2                 ; Implemented" bit

;       xor     ax,ax                   ; Disable A20. Fixes PS2 Ctl-Alt-Del bug
;       call    PS2_A20Handler
@@NoPS2:                                ;
        ret                             ;
        endp                            ;

;---------------------------------------------------------------------------;
;                                                                           ;
;  IsHPMachine  - Check for HP Vectra Machine                               ;
;                                                                           ;
;  Expects: None                                                            ;
;  Returns: NC if we're on a HP Vectra machine, CY otherwise                ;
;  Regs:    AX, ES and Flags clobbered                                      ;
;                                                                           ;
;---------------------------------------------------------------------------;

        proc    IsHPMachine

        mov     ax,0F000h               ;
        mov     es,ax                   ;
        cmp     [word es:0F8h],'PH'     ;
        je      @@FoundIt               ;
        stc                             ;
        ret                             ;
@@FoundIt:                              ;
        clc                             ;
        ret                             ;
        endp                            ;
ends
        end                             ; End Of Module
